/*
 * Copyright (C) 2006  Chris Rose
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 */

package com.aimedia.azjabber.xmpp;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.jivesoftware.smack.ConnectionListener;
import org.jivesoftware.smack.PacketListener;
import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.XMPPException;
import org.jivesoftware.smack.filter.PacketTypeFilter;
import org.jivesoftware.smack.packet.Message;
import org.jivesoftware.smack.packet.Packet;
import org.jivesoftware.smack.packet.Presence;
import org.jivesoftware.smack.packet.Presence.Type;

import com.aimedia.azjabber.AzJabberPlugin;
import com.aimedia.azjabber.Logger;
import com.aimedia.azjabber.config.IAzJabberMessageKeys;
import com.aimedia.azjabber.config.JabberConfiguration;
import com.aimedia.azjabber.config.JabberConnectionParameters;
import com.aimedia.azjabber.services.ICommunicationServiceProvider;
import com.aimedia.azjabber.services.IServiceConnectionEventListener;

public class XMPPConnectionManager implements ICommunicationServiceProvider {

    private static XMPPConnectionManager _instance;

    private boolean connected = false;
    
    public static final ICommunicationServiceProvider instance () {
        if (_instance == null) {
            _instance = new XMPPConnectionManager ();
        }
        return _instance;
    }

    private XMPPConnection                     connection;

    private List<IServiceConnectionEventListener> listeners = Collections
                                                                 .synchronizedList (new ArrayList<IServiceConnectionEventListener> ());

    private JabberConfiguration                config;

    private ConnectionListenerImpl cl;

    private IncomingPacketListenerImpl pl;

    private OutgoingPacketListenerImpl pwl;

    private XMPPConnectionManager () {
        config = AzJabberPlugin.getPlugin ().getConfiguration ();
    }

    public boolean isConnected () {
        return connected && connection != null && connection.isConnected ();
    }

    public void disconnect () {
        if (connection != null) {
            Logger.info (IAzJabberMessageKeys.LOG_DISCONNECTING, new String[] { connection.getServiceName () });
            connection.removeConnectionListener (cl);
            connection.removePacketListener (pl);
            connection.removePacketWriterListener (pwl);
            connection.close ();
            notifyListenersOfDisconnection ();
            connected = false;
        }
    }

    public void connect (JabberConnectionParameters connectionInformation) throws XMPPException {
        if (connection != null && connection.isConnected ()) {
            // : reconnect physically
            disconnect ();
        }
        connection = new XMPPConnection (connectionInformation.getConnectionConfiguration ());
        cl = new ConnectionListenerImpl();
        pl = new IncomingPacketListenerImpl ();
        pwl = new OutgoingPacketListenerImpl ();
        connection.addConnectionListener (cl);
        connection.addPacketListener (pl, new PacketTypeFilter (Presence.class));
        connection.addPacketWriterListener (pwl, null);
        Logger.info (IAzJabberMessageKeys.LOG_CONNECTED, new String[] { connectionInformation.getHost () });

        // : log in
        connection.login (connectionInformation.getUsername (), new String (connectionInformation.getPassword ()),
                connectionInformation.getResourceBase (), true);
        Logger.info (IAzJabberMessageKeys.LOG_AUTHENTICATED, new String[] { connectionInformation.getUsername (),
                connection.getServiceName () });

        // : TODO determine if it's possible to set invisible right after
        // connecting and still get presence packets.
        connected = true;
    }

    private void notifyListeners (String JID, Presence.Type presenceType) {
        for (IServiceConnectionEventListener listener : listeners) {
            if (presenceType == Type.AVAILABLE) {
                listener.recipientOnline (JID);
            }
            else if (presenceType == Type.UNAVAILABLE) {
                listener.recipientOffline (JID);
            }
            else {
                Logger.info (IAzJabberMessageKeys.LOG_NOT_HANDLED_PRESENCE, new Object[] { presenceType });
            }
        }
    }
    
    private void notifyListenersOfDisconnection () {
        for (IServiceConnectionEventListener listener : listeners) {
            listener.serviceOffline (this);
        }
    }

    private class OutgoingPacketListenerImpl implements PacketListener {

        public void processPacket (Packet packet) {
            if (config.isLoggingDetailed ()) {
                Logger.info (IAzJabberMessageKeys.LOG_SENT_PACKET, packet.toXML ());
            }
        }

    }

    private class IncomingPacketListenerImpl implements PacketListener {

        public void processPacket (Packet packet) {
            assert (packet instanceof Presence);
            Presence p = (Presence) packet;
            Logger.info (IAzJabberMessageKeys.LOG_RECEIVED_PRESENCE, p.getFrom ());

            if (p.getFrom () != null && p.getType () != Presence.Type.ERROR) {
                notifyListeners (p.getFrom (), p.getType ());
            }
            
            if (p.getType () == Presence.Type.ERROR) {
                Logger.warn (IAzJabberMessageKeys.LOG_PRESENCE_ERROR, new Object[] {p.getFrom (), (p.getError () == null ? p.toXML () : p.getError ().toXML ())});
            }
        }
    }

    private class ConnectionListenerImpl implements ConnectionListener {

        public void connectionClosed () {
        }

        public void connectionClosedOnError (Exception e) {
            disconnect ();
            Logger.error (IAzJabberMessageKeys.LOG_CONNECTION_FAILED, e);
        }

    }

    public void sendMessage (String to, String eventMessage) {
        Exception connectionError = null;
        if (! connected) { 
            try {
                connect (config.getJabberConnectionParameters ());
            }
            catch (XMPPException e) {
                connectionError = e;
            }
        }
        if (isConnected ()) {
            Message message = new Message (to);
            message.setBody (eventMessage);
            connection.sendPacket (message);
        }
        else {
            Logger.alert (IAzJabberMessageKeys.LOG_FAILED_TO_SEND_MESSAGE, new Object[] {to, eventMessage}, connectionError);
        }
    }

    public void addServiceConnectionEventListener (IServiceConnectionEventListener listener) {
        this.listeners.add (listener);
    }

    public void removeServiceConnectionEventListener (IServiceConnectionEventListener listener) {
        this.listeners.remove (listener);
    }

    public boolean checkConnection () {
        if (! connected) {
            try {
                connect (config.getJabberConnectionParameters ());
            }
            catch (XMPPException e) {
                Logger.error (IAzJabberMessageKeys.LOG_CONNECTION_FAILED, e);
            }
        }
        
        return connected;
    }

}
